/*******************************************************************************
rpc.c

Copyright (C) 2000 Simon Nieuviarts

An implementation of the RPC.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version. 

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*******************************************************************************/
/* TODO :
- Set locking everywhere.
- Verify the reference counters on the objects.
- Verify the error numbers chosen.
- Do a better authentication handling.
- Study the removal of the active wait in do_rpc_call, or the possibility of
user's interrupt.
*/

#include <oskit/dev/osenv_mem.h>
#include <oskit/dev/osenv_timer.h>
#include <oskit/net/socket.h>
#include <oskit/com/stream.h>
#include <oskit/io/asyncio.h>

#include <oskit/c/string.h>
#include <oskit/c/netinet/in.h>
#include <oskit/c/arpa/inet.h>
#include <oskit/c/assert.h>


#include "rpc.h"


#define PROG_PORTMAP 100000
#define PORTMAP_LOOKUP 3
#define PORTMAP_PORT 111


/* The state of the reply to an RPC call. */
typedef enum oskit_rpc_reply_stat
{
    OSKIT_RPC_SUCCESS = 0,
    OSKIT_RPC_PROG_UNAVAIL = 1,
    OSKIT_RPC_PROG_MISMATCH = 2,
    OSKIT_RPC_PROC_UNAVAIL = 3,
    OSKIT_RPC_GARBAGE_ARGS = 4,
    OSKIT_RPC_SYSTEM_ERR = 5,
    OSKIT_RPC_RPC_MISMATCH = 100,
    OSKIT_RPC_AUTH_ERROR = 101
} oskit_rpc_reply_stat ;


/* The XID counter is global. */
#include <oskit/dev/osenv_rtc.h>
static oskit_u32_t c_xid = 0 ;


oskit_u32_t
get_new_xid( void )
{
    oskit_u32_t xid = c_xid ;

    /* XXX Locking needed here. */
    if ( c_xid == 0 )
    {
        oskit_timespec_t tod ;
        oskit_osenv_rtc_t *rtc ;
        oskit_error_t err ;

        c_xid = 1 ;
        rtc = oskit_create_osenv_rtc() ;
        if ( rtc == 0 ) goto return_xid ;
        err = oskit_osenv_rtc_get( rtc, &tod ) ;
        if ( OSKIT_FAILED( err ) ) goto return_xid ;
        c_xid = tod.tv_sec << 12 ;  /* XXX to explain (same as Linux). */
        xid = c_xid ;
    }
    c_xid++ ;

return_xid:
    /* XXX Unlocking needed here. */
    return xid ;
}


/* The number of RPC objects using this timer. */
static unsigned int timer_ref_counter = 0 ;
static oskit_osenv_timer_t *osenv_timer ;


#define TIMER_FREQ 100
#define TIME_DIV (1000/TIMER_FREQ)
static volatile oskit_u32_t ticks = 0 ;


static void
timer_interrupt( void )
{
    ticks++ ;
}


static inline oskit_u32_t
get_ticks( void )
{
    return ticks ;
}


static struct oskit_rpc_ops rpc_ops ;


/* Create an object to do RPC calls. */
oskit_error_t
oskit_rpc_init( oskit_osenv_mem_t *mem,
                oskit_osenv_timer_t *timer,
                oskit_socket_factory_t *sf,
                const char *localhost,
                oskit_u32_t short_timeout,
                oskit_u32_t long_timeout,
                oskit_rpc_t **out_rpc )
{
    oskit_rpc_t *rpc ;
    char *host ;

    /* Allocate memory for the RPC object. */
    rpc = oskit_osenv_mem_alloc( mem, sizeof(*rpc), OSENV_NONBLOCKING, 0 ) ;
    if ( rpc == 0 )
    {
        oskit_osenv_mem_release( mem ) ;
        oskit_osenv_timer_release( timer ) ;
        oskit_socket_factory_release( sf ) ;
        return OSKIT_ENOMEM ;
    }

    /* Allocate memory for the host name. */
    host = oskit_osenv_mem_alloc( mem, strlen(localhost) + 1,
                                  OSENV_NONBLOCKING, 0 ) ;
    if ( host == 0 )
    {
        oskit_osenv_mem_free( mem, rpc, OSENV_NONBLOCKING, sizeof(*rpc) ) ;
        oskit_osenv_mem_release( mem ) ;
        oskit_osenv_timer_release( timer ) ;
        oskit_socket_factory_release( sf ) ;
        return OSKIT_ENOMEM ;
    }
    strcpy( host, localhost ) ;

    /* Setup global timer or add a reference on it. */
    if ( timer_ref_counter == 0 )
    {
        oskit_osenv_timer_register( timer, timer_interrupt, TIMER_FREQ ) ;   
        osenv_timer = timer ;
    }
    else
        oskit_osenv_timer_release( timer ) ;
    timer_ref_counter++ ;

    rpc->ops = &rpc_ops ;
    rpc->s = 0 ;
    rpc->sf = sf ;
    rpc->mem = mem ;
    rpc->ref_counter = 1 ;
    rpc->localhost = host ;
    rpc->short_timeout = short_timeout ;
    rpc->long_timeout = long_timeout ;
    rpc->prog = 0 ;
    rpc->vers = 0 ;
    rpc->wptr = 0 ;
    rpc->rptr = 0 ;
    rpc->rlen = 0 ;

    *out_rpc = rpc ;
    return 0 ;
}


static OSKIT_COMDECL 
rpc_query( oskit_rpc_t *s, const oskit_iid_t *iid, void **out_ihandle )
{
    if ( ( memcmp( iid, &oskit_iunknown_iid, sizeof(*iid) ) == 0 ) ||
         ( memcmp( iid, &oskit_rpc_iid, sizeof(*iid) ) == 0 ) )
    {
        *out_ihandle = s ;
        return 0 ;
    }

    *out_ihandle = 0 ;
    return OSKIT_E_NOINTERFACE ;
}


static OSKIT_COMDECL_U
rpc_addref( oskit_rpc_t *rpc )
{
    return ++(rpc->ref_counter) ;
}


static OSKIT_COMDECL_U
rpc_release( oskit_rpc_t *rpc )
{
    oskit_osenv_mem_t *mem = rpc->mem ;

    if ( (--rpc->ref_counter) == 0 )
    {
        /* Release the referenced objects and free the allocated memory. */
        if ( rpc->s ) oskit_stream_release( rpc->s ) ;
        oskit_socket_factory_release( rpc->sf ) ;
        oskit_osenv_mem_free( mem, rpc, OSENV_NONBLOCKING, sizeof(*rpc) ) ;
        oskit_osenv_mem_release( mem ) ;

        /* Release one reference on the global timer. */
        if ( (--timer_ref_counter) == 0 )
            oskit_osenv_timer_release( osenv_timer ) ;
        return 0 ;
    }
    else
        return rpc->ref_counter ;
}


/* Write an int in the XDR format in BUF and returns a pointer after the int. */
static char *
wxdr_int( char *buf, int i )
{
    *buf++ = i >> 24 ;
    *buf++ = i >> 16 ;
    *buf++ = i >> 8 ;
    *buf++ = i ;
    return buf ;
}


/* Write an int in the args of an initialized call. */
static OSKIT_COMDECL
rpc_call_wxdr_int( oskit_rpc_t *rpc, oskit_u32_t x )
{
    if ( rpc->wptr == 0 )
        /* Call not initialized. */
        /* XXX A better error number ? */
        return OSKIT_E_FAIL ;

    if ( ( rpc->wptr + 4 - rpc->call_buf ) > RPC_BUF_LEN )
        /* Not enough room in buffer. */
        return OSKIT_E_FAIL ;

    rpc->wptr = wxdr_int( rpc->wptr, x ) ;
    return 0 ;
}


/* Write an opaque object in the args of an initialized call. */
static OSKIT_COMDECL
rpc_call_wxdr_opaque( oskit_rpc_t *rpc, const void *buf, oskit_size_t l )
{
    char *p ;

    if ( rpc->wptr == 0 )
        /* Call not initialized. */
        /* XXX A better error number ? */
        return OSKIT_E_FAIL ;

    if ( ( rpc->wptr + 4 + l - rpc->call_buf ) > RPC_BUF_LEN )
        /* Not enough room in buffer. */
        return OSKIT_E_FAIL ;

    p = wxdr_int( rpc->wptr, l ) ;
    memcpy( p, buf, l ) ;
    while ( ( l & 3 ) != 0 ) p[l++] = 0 ;
    rpc->wptr = p + l ;

    return 0 ;
}


/* Write a string in the args of an initialized call. */
static OSKIT_COMDECL
rpc_call_wxdr_string( oskit_rpc_t *rpc, const char *string )
{
    return rpc_call_wxdr_opaque( rpc, string, strlen(string) ) ;
}


/* Write data in the args of an initialized call. */
static OSKIT_COMDECL
rpc_call_wxdr_data( oskit_rpc_t *rpc, const void *buf, oskit_size_t l )
{
    if ( rpc->wptr == 0 )
        /* Call not initialized. */
        /* XXX A better error number ? */
        return OSKIT_E_FAIL ;

    if ( ( rpc->wptr + l - rpc->call_buf ) > RPC_BUF_LEN )
        /* Not enough room in buffer. */
        return OSKIT_E_FAIL ;

    memcpy( rpc->wptr, buf, l ) ;
    while ( ( l & 3 ) != 0 ) rpc->wptr[l++] = 0 ;
    rpc->wptr += l ;

    return 0 ;
}


/* Write a string in the XDR format in BUF and returns a pointer after the
string. */
static char *
wxdr_string( char *buf, char *s )
{
    unsigned int l = strlen( s ) ;

    if ( l > 255 ) l = 255 ;
    buf = wxdr_int( buf, l ) ;
    memcpy( buf, s, l ) ;
    while ( ( l & 3 ) != 0 ) buf[l++] = 0 ;
    return buf + l ;
}


#if 0
/* Initialize a call of the procedure PROC of a connected RPC object. */
static OSKIT_COMDECL
rpc_call_init_no_auth( oskit_rpc_t *rpc, oskit_u32_t proc )
{
    char *p ;

    /* XXX A better error number ? */
    if ( rpc->s == 0 ) return OSKIT_E_FAIL ;

/*
The RPC CALL message structure format (without authentication) :
0: unsigned int xid ;
4: int mtype = 0 ;
8: unsigned int rpcvers = 2 ;
12: unsigned int prog ;
16: unsigned int vers ;
20: unsigned int proc ;
24: opaque_auth cred = 0, 0 ;
32: opaque_auth verf = 0, 0 ;
40: Procedure parameters.
*/
    p = wxdr_int( rpc->call_buf, get_new_xid() ) ;
//XXX    p = wxdr_int( rpc->call_buf, rpc->c_xid++ ) ;
    p = wxdr_int( p, 0 ) ;
    p = wxdr_int( p, 2 ) ;
    p = wxdr_int( p, rpc->prog ) ;
    p = wxdr_int( p, rpc->vers ) ;
    p = wxdr_int( p, proc ) ;
    p = wxdr_int( p, 0 ) ;
    p = wxdr_int( p, 0 ) ;
    p = wxdr_int( p, 0 ) ;
    p = wxdr_int( p, 0 ) ;
    rpc->wptr = p ;

    return 0 ;
}
#endif


/* Initialize a call of the procedure PROC of a connected RPC object. */
static OSKIT_COMDECL
rpc_call_init( oskit_rpc_t *rpc, oskit_u32_t proc )
{
    char *p ;
    char *p2 ;

    /* XXX A better error number ? */
    if ( rpc->s == 0 ) return OSKIT_E_FAIL ;

    p = wxdr_int( rpc->call_buf, get_new_xid() ) ;
//XXX    p = wxdr_int( rpc->call_buf, rpc->c_xid++ ) ;
    p = wxdr_int( p, 0 ) ;
    p = wxdr_int( p, 2 ) ;
    p = wxdr_int( p, rpc->prog ) ;
    p = wxdr_int( p, rpc->vers ) ;
    p = wxdr_int( p, proc ) ;
    p = wxdr_int( p, 1 ) ; /* AUTH_SYS */
    p2 = p ;
    p += 4 ;
    p = wxdr_int( p, 0 ) ; /* stamp */
    p = wxdr_string( p, rpc->localhost ) ; /* machinename */
    p = wxdr_int( p, 0 ) ; /* uid */
    p = wxdr_int( p, 0 ) ; /* gid */
    p = wxdr_int( p, 1 ) ; /* nb gids */
    p = wxdr_int( p, 0 ) ; /* gids[0] */
    wxdr_int( p2, p-(p2+4) ) ; /* auth. size */
    p = wxdr_int( p, 0 ) ;
    p = wxdr_int( p, 0 ) ;
    rpc->wptr = p ;

    return 0 ;
}


/* Read an int in the XDR format in buf. */
static OSKIT_COMDECL
rpc_reply_rxdr_int( oskit_rpc_t *rpc, oskit_u32_t *x )
{
    oskit_u32_t i ;
    char *p = rpc->rptr ;

    if ( rpc->rlen < 4 ) return OSKIT_E_FAIL ;

    i = (unsigned char)*p++ ;
    i = ( i << 8 ) + (unsigned char)*p++ ;
    i = ( i << 8 ) + (unsigned char)*p++ ;
    i = ( i << 8 ) + (unsigned char)*p ;
    *x = i ;
    rpc->rptr += 4 ;
    rpc->rlen -= 4 ;

    return 0 ;
}


/* Read an opaque object or a string. */
static OSKIT_COMDECL
rpc_reply_rxdr_opaque( oskit_rpc_t *rpc, char **p, oskit_size_t *l )
{
    oskit_error_t err ;
    oskit_size_t s ;
    oskit_size_t s2 ;

    err = rpc_reply_rxdr_int( rpc, &s ) ;
    if ( OSKIT_FAILED( err ) ) return err ;
    s2 = ( s + 3 ) & (~3) ;
    if ( rpc->rlen < s2 ) return OSKIT_E_FAIL ;

    *l = s ;
    *p = rpc->rptr ;
    rpc->rptr += s2 ;
    rpc->rlen -= s2 ;
    return 0 ;
}


/* Read a well defined length of data (rounded to a multiple of 4). */
static OSKIT_COMDECL
rpc_reply_rxdr_data( oskit_rpc_t *rpc, void **p, oskit_size_t l )
{
    l = ( l + 3 ) & (~3) ;
    if ( rpc->rlen < l ) return OSKIT_E_FAIL ;

    *p = rpc->rptr ;
    rpc->rptr += l ;
    rpc->rlen -= l ;
    return 0 ;
}


/* Read a well defined length of data (rounded to a multiple of 4). */
static OSKIT_COMDECL_U
rpc_reply_len( oskit_rpc_t *rpc )
{
    return rpc->rlen ;
}


/* Read an unsigned int in the XDR format in buf. */
static unsigned int
rxdr_unsigned_int( const char *buf )
{
    unsigned int i ;

    i = (unsigned char)*buf++ ;
    i = ( i << 8 ) + (unsigned char)*buf++ ;
    i = ( i << 8 ) + (unsigned char)*buf++ ;
    i = ( i << 8 ) + (unsigned char)*buf ;

    return i ;
}


/* Convert an RPC reply_state to a system error number. */
static oskit_error_t
rpc_to_oskit_error( oskit_rpc_reply_stat r )
{
    switch( r )
    {
    case OSKIT_RPC_SUCCESS:
        return 0 ;
    case OSKIT_RPC_PROG_UNAVAIL:
        return OSKIT_ESRCH ;
    case OSKIT_RPC_PROG_MISMATCH:
    case OSKIT_RPC_PROC_UNAVAIL:
        return OSKIT_ENOTSUP ;
    case OSKIT_RPC_SYSTEM_ERR:
        return OSKIT_ENOMEM ;
    case OSKIT_RPC_AUTH_ERROR:
        return OSKIT_EACCES ;
    case OSKIT_RPC_GARBAGE_ARGS:
    case OSKIT_RPC_RPC_MISMATCH:
    default:
        return OSKIT_EINVAL ;
    }
}


/* Do the RPC call. */
static OSKIT_COMDECL
rpc_do_call( oskit_rpc_t *rpc )
{
    oskit_error_t err ;
    oskit_size_t l ;
    oskit_rpc_reply_stat st ;
    char *p ;
    oskit_u32_t t1, t2 ;
    oskit_asyncio_t *io ;

    err = oskit_stream_query( rpc->s, &oskit_asyncio_iid, (void **)&io ) ;
    if ( OSKIT_FAILED( err ) ) return err ;

    /* For the long timeout. */
    t1 = get_ticks() ;

timeout_retry:
    t2 = get_ticks() ;
    if ( (t2-t1) > (rpc->long_timeout+TIME_DIV-1)/TIME_DIV )
    {
        err = OSKIT_ETIMEDOUT ;
        goto return_error ;
    }

    err = oskit_stream_write( rpc->s, rpc->call_buf,
                              rpc->wptr-rpc->call_buf, &l ) ;
    if ( OSKIT_FAILED( err ) ) goto return_error ;

wait_to_read:
    do
    {
        err = oskit_asyncio_poll( io ) ;
        if ( OSKIT_FAILED( err ) ) goto return_error ;
        if ( ( err & (OSKIT_ASYNCIO_READABLE|OSKIT_ASYNCIO_EXCEPTION) ) )
            goto something_to_read ;
    }
    while( (get_ticks()-t2) < ((rpc->short_timeout+TIME_DIV-1)/TIME_DIV) ) ;
    goto timeout_retry ;

something_to_read:
    p = rpc->reply_buf ;
    err = oskit_stream_read( rpc->s, p, RPC_BUF_LEN, &l ) ;
    if ( OSKIT_FAILED( err ) )
    {
        /* XXX There should be a better error analysis here. */
        goto return_error ;
    }

    if ( l < 12 ) goto wait_to_read ;
    /* Compare the xid with the one of the emitted call. */
    if ( rxdr_unsigned_int( rpc->call_buf ) !=
         rxdr_unsigned_int( p ) ) goto wait_to_read ;
    if ( rxdr_unsigned_int( p + 4 ) != 1 ) goto wait_to_read ;
    if ( rxdr_unsigned_int( p + 8 ) == 0 )
    {
        /* Message accepted. */
        int i ;

        l -= 12 ;
        p += 12 ;
        /* Read the authentication verifier. */
        if ( l < 8 ) goto wait_to_read ;

        /* XXX AUTH_NONE awaited : to change. */
        if ( rxdr_unsigned_int( p ) != 0 ) assert(0) ;
        i = ( rxdr_unsigned_int( p + 4 ) + 3 ) & (~3) ;
        p += i + 8 ; l -= i + 8 ;
        if ( l < 4 ) goto wait_to_read ;
        st = rxdr_unsigned_int( p ) ;
    }
    else
    {
        /* Message refused. */
        p += 12 ;
        if ( l < 4 ) goto wait_to_read ;
        st = rxdr_unsigned_int( p ) + 100 ;
    }

    rpc->rptr = p + 4 ;
    rpc->rlen = l - 4 ;
    err = rpc_to_oskit_error( st ) ;

return_error:
    oskit_asyncio_release( io ) ;
    return err ;
}


/* Initialize a stream *S to do RPC on UDP to host HOST on port RP. The stream
must be released after usage. */
static oskit_error_t
init_rpc_stream( const char *host, unsigned short rp,
                 oskit_socket_factory_t *sf, oskit_stream_t **s )
{
    struct sockaddr_in addr ;
    unsigned short lp ;
    oskit_socket_t *so;
    oskit_error_t err;

    err = oskit_socket_factory_create( sf, OSKIT_AF_INET, OSKIT_SOCK_DGRAM,
                                       IPPROTO_UDP, &so ) ;
    if ( err ) return err ;
    
    /* The local address. */
    for( lp=620; lp<1000; lp++ )
    {
        memset( &addr, 0, sizeof(addr) ) ;
        addr.sin_addr.s_addr = INADDR_ANY ;
        addr.sin_family = OSKIT_AF_INET ;
        addr.sin_port = htons( lp ) ;
        err = oskit_socket_bind( so, (struct oskit_sockaddr *)&addr,
                                 sizeof(addr) ) ;
	if ( err == 0 ) goto bound ;
        if ( err != OSKIT_EADDRINUSE )
        {
	    oskit_socket_release( so ) ;
            return err ;
        }
    }
    oskit_socket_release( so ) ;
    return err ;

bound:
    /* The remote address. */
    memset( &addr, 0, sizeof(addr) ) ;
    if ( inet_aton( host, &addr.sin_addr ) == 0 )
    {
        oskit_socket_release( so ) ;
        return OSKIT_EINVAL ;
    }
    addr.sin_family = OSKIT_AF_INET ;
    addr.sin_port = htons( rp ) ;

    err = oskit_socket_connect( so, (struct oskit_sockaddr *)&addr, sizeof(addr) ) ;
    if ( err )
    {
        oskit_socket_release( so ) ;
        return err ;
    }

    err = oskit_socket_query( so, &oskit_stream_iid, (void**)s ) ;
    oskit_socket_release( so ) ;
    return err ;
}


/* Connect the RPC object to the RPC program PROG version VERS on HOST. */
static OSKIT_COMDECL
rpc_connect( oskit_rpc_t *rpc, const char *host,
             oskit_u32_t prog, oskit_u32_t vers )
{
    oskit_error_t err ;
    oskit_u32_t r ;

    /*  Eventually close an open stream. */
    if ( rpc->s ) oskit_stream_release( rpc->s ) ;

    /* Initialize the stream to the server. */
    err = init_rpc_stream( host, PORTMAP_PORT, rpc->sf, &rpc->s ) ;
    if ( OSKIT_FAILED( err ) ) return err ;
    rpc->prog = PROG_PORTMAP ;
    rpc->vers = 2 ;

    /* Build the body of the RPC call. No error to check. */
    rpc_call_init( rpc, PORTMAP_LOOKUP ) ;
    rpc_call_wxdr_int( rpc, prog ) ;
    rpc_call_wxdr_int( rpc, vers ) ;
    rpc_call_wxdr_int( rpc, IPPROTO_UDP ) ;
    rpc_call_wxdr_int( rpc, 0 ) ;

    /* Do the call. */
    err = rpc_do_call( rpc ) ;
    oskit_stream_release( rpc->s ) ;
    rpc->s = 0 ;
    if ( OSKIT_FAILED( err ) ) return err ;

    /* The service exists. */
    err = rpc_reply_rxdr_int( rpc, &r ) ;
    if ( OSKIT_FAILED( err ) ) return err ;
    /* r = service port no. */

    /* Create the stream to the service. */
    err = init_rpc_stream( host, r, rpc->sf, &rpc->s ) ;
    if ( OSKIT_FAILED( err ) ) return err ;

    rpc->prog = prog ;
    rpc->vers = vers ;
    return 0 ;
}


static struct oskit_rpc_ops rpc_ops =
{
    rpc_query,
    rpc_addref,
    rpc_release,
    rpc_call_init,
    rpc_call_wxdr_int,
    rpc_call_wxdr_opaque,
    rpc_call_wxdr_string,
    rpc_call_wxdr_data,
    rpc_reply_rxdr_int,
    rpc_reply_rxdr_opaque,
    rpc_reply_rxdr_data,
    rpc_reply_len,
    rpc_do_call,
    rpc_connect
} ;
